"""
Plugins
=======

This package contains the various built-in plugins available for Chancy.
Plugins are used to extend the functionality of Chancy, while allowing for
significant customization.

Chancy comes with several built-in plugins:

- :class:`chancy.plugins.pruner.Pruner`: This plugin is responsible for pruning
    old jobs from the database.
- :class:`chancy.plugins.recovery.Recovery`: This plugin is responsible for
    recovering jobs that were interrupted during execution.
- :class:`chancy.plugins.leadership.Leadership`: This plugin is responsible for
    managing leadership elections between multiple workers.
- :class:`chancy.plugins.cron.Cron`: This plugin is responsible for scheduling
    jobs to run at specific times.
- :class:`chancy.plugins.api.Api`: This plugin provides an API and dashboard
    for viewing the state of the Chancy cluster.
- :class:`chancy.plugins.workflow.WorkflowPlugin`: This plugin provides a
    workflow system for managing complex job dependencies.
- :class:`chancy.plugins.retry.RetryPlugin`: This plugin provides a slightly
    more advanced job retry system with jitter, backoff, and other features.
"""



import asyncio
import datetime
import enum
import inspect
import uuid
import time
import json
import secrets
import itertools
import contextlib


@contextlib.contextmanager
def timed_block():
    """
    A context manager that times the block of code within it. The elapsed
    time will be updated in the `elapsed` attribute of the returned object
    until the block is exited.

    Example:

    .. code-block:: python

        with timed_block() as timer:
            time.sleep(1)

        print(f"Block took {timer.elapsed} seconds.")
    """

    class Timer:
        def __init__(self):
            self.start = time.monotonic()
            self.stop = None

        @property
        def elapsed(self):
            return (self.stop or time.monotonic()) - self.start

    timer = Timer()
    yield timer
    timer.stop = time.monotonic()


def importable_name(obj):
    """
    Get the importable name for an object.

    .. note::

        This will only work for objects that are actually importable,
        i.e. they are defined in a module. Lambdas, for example, will not
        have an importable name.

    :param obj: The object to get the importable name for.
    :return: str
    """
    if obj.__class__.__module__ == "__main__":
        module = inspect.getmodule(obj)
        return f"{module.__spec__.name}.{obj.__qualname__}"

    return f"{obj.__module__}.{obj.__qualname__}"


def import_string(name):
    """
    Import an object from a string previously created using
    :func:`importable_name`.

    :param name: The importable name of the object.
    :return: Any
    """
    mod_name, _, func_name = name.rpartition(".")
    mod = __import__(mod_name, fromlist=[func_name])
    return getattr(mod, func_name)


def chancy_uuid() -> str:
    """
    Generate a UUID suitable for use as a job ID.

    .. note::

        It's UUID7, kinda, since the draft keeps changing.

    :return: str
    """
    t = (time.time_ns() // 100) & 0xFFFFFFFFFFFFFF
    rand = secrets.randbits(62)
    uuid = (t << 68) | (7 << 64) | (2 << 62) | rand
    return f"{uuid:032x}"


def json_dumps(obj, **kwargs):
    """
    Serialize an object to a JSON formatted str with support for UUIDs.

    :param obj: The object to serialize.
    :param kwargs: Additional arguments to pass to `json.dumps`.
    :return: str
    """

    def _dump(o):
        if isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, datetime.datetime):
            return o.isoformat()
        elif isinstance(o, enum.Enum):
            return o.value

        raise TypeError(
            f"Object of type {o.__class__.__name__} is not JSON serializable"
        )

    return json.dumps(obj, default=_dump, **kwargs)


def chunked(iterable, size):
    """
    Yield chunks of `size` from `iterable`.

    :param iterable: The iterable to chunk.
    :param size: The size of each chunk.
    """
    it = iter(iterable)
    while True:
        chunk = list(itertools.islice(it, size))
        if not chunk:
            break
        yield chunk


class TaskManager:
    """
    A simple task manager that keeps track of tasks.

    The asyncio.TaskGroup is not used here for a couple of reasons, but mostly
    because it's only available in Python 3.11+ and we want to maintain
    compatibility with earlier versions.
    """

    def __init__(self):
        self.tasks: set[asyncio.Task] = set()

    def add(self, coro, *, name: str | None = None) -> asyncio.Task:
        """
        Add a task to the manager.
        """
        task = asyncio.create_task(coro)
        task.add_done_callback(self.tasks.remove)
        self.tasks.add(task)
        if name:
            task.set_name(name)
        return task

    def remove(self, task: asyncio.Task):
        """
        Remove a task from the manager.
        """
        task.result()
        task.cancel()
        self.tasks.discard(task)

    async def run(self):
        """
        Wait for all tasks to complete.
        """
        while self.tasks:
            await asyncio.wait(self.tasks, return_when=asyncio.ALL_COMPLETED)

    async def shutdown(self, *, timeout: int | None = None) -> bool:
        """
        Cancel all tasks and wait for them to complete.

        :param timeout: The number of seconds to wait for a clean shutdown
                        before forcing the tasks to stop.
        :return: False if a timeout occurred trying to cancel the tasks, True
                 if shutdown was successful.
        """
        for task in self.tasks:
            task.cancel()

        try:
            await asyncio.wait_for(
                asyncio.gather(*self.tasks, return_exceptions=True),
                timeout=timeout,
            )
        except (asyncio.TimeoutError, TimeoutError):
            self.tasks = {task for task in self.tasks if not task.done()}
            return False
        except asyncio.CancelledError:
            pass

        self.tasks.clear()
        return True



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "chancy"
copyright = "2024, Tyler Kennedy"
author = "Tyler Kennedy"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    "sphinx.ext.graphviz",
    "sphinx.ext.linkcode",
    "sphinx_inline_tabs",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

language = "en"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_logo = "../misc/logo.png"
html_title = "Chancy"

# -- Options for todo extension ----------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/todo.html#configuration

todo_include_todos = True


def linkcode_resolve(domain, info):
    if domain != "py":
        return None

    if not info["module"]:
        return None

    filename = info["module"].replace(".", "/")
    return f"https://github.com/tktech/chancy/blob/main/{filename}.py"



import asyncio
import secrets
from typing import AsyncIterator

import pytest
import pytest_asyncio
from pytest_postgresql import factories

from chancy import Chancy, Worker


external_postgres = factories.postgresql_noproc(
    host="localhost",
    password="localtest",
    user="postgres",
    port=8190,
    dbname=f"chancy_test_{secrets.token_hex(8)}",
)
postgresql = factories.postgresql(
    "external_postgres",
)


@pytest_asyncio.fixture()
async def chancy(request, postgresql):
    """
    Provides a Chancy application instance with an open connection pool
    to the test database.
    """
    i = postgresql.info

    async with Chancy(
        dsn=f"postgresql://{i.user}:{i.password}@{i.host}:{i.port}/{i.dbname}",
        **getattr(request, "param", {}),
    ) as chancy:
        await chancy.migrate()
        yield chancy
        await chancy.migrate(to_version=0)


@pytest.fixture
def chancy_just_app(postgresql):
    """
    Provides just a configured chancy instance with no open connection pool
    or migrations.
    """
    i = postgresql.info
    return Chancy(
        dsn=f"postgresql://{i.user}:{i.password}@{i.host}:{i.port}/{i.dbname}",
    )


@pytest_asyncio.fixture()
async def worker(chancy) -> AsyncIterator[tuple[Worker, asyncio.Task]]:
    """
    Starts and returns a Worker and the task associated with it.

    If the worker is not stopped by the time the test completes, it will be
    cancelled.
    """
    worker = Worker(chancy)
    worker_task = asyncio.create_task(worker.start())

    try:
        yield worker, worker_task
    finally:
        if not worker_task.done():
            worker_task.cancel()

        try:
            await asyncio.gather(worker_task)
        except asyncio.CancelledError:
            pass


@pytest_asyncio.fixture()
async def worker_no_start(chancy) -> Worker:
    """
    Returns a Worker instance that has not been started.
    """
    return Worker(chancy)



import asyncio

import pytest

from chancy import Chancy, Worker, Queue, Job, QueuedJob


def job_to_run():
    pass


async def async_job_to_run():
    pass


def job_with_instance(*, job: QueuedJob):
    job.meta["received_instance"] = True


async def async_job_with_instance(*, job: QueuedJob):
    job.meta["received_instance"] = True


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "executor",
    [
        "chancy.executors.process.ProcessExecutor",
        "chancy.executors.thread.ThreadedExecutor",
    ],
)
async def test_basic_job_sync(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], executor: str
):
    """
    Ensures that a basic job can be run on all built-in executors.
    """
    await chancy.declare(Queue("low", executor=executor))

    ref = await chancy.push(Job.from_func(job_to_run, queue="low"))
    job = await chancy.wait_for_job(ref)

    assert job.state == job.State.SUCCEEDED


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "executor",
    [
        "chancy.executors.asyncex.AsyncExecutor",
    ],
)
async def test_basic_job_async(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], executor: str
):
    """
    Ensures that a basic job can be run on all built-in executors.
    """
    await chancy.declare(Queue("low", executor=executor))

    ref = await chancy.push(Job.from_func(async_job_to_run, queue="low"))
    job = await chancy.wait_for_job(ref)

    assert job.state == job.State.SUCCEEDED


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "executor",
    [
        "chancy.executors.process.ProcessExecutor",
        "chancy.executors.thread.ThreadedExecutor",
    ],
)
async def test_job_instance_kwarg(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], executor: str
):
    """
    Test that jobs requesting a QueuedJob kwarg receive the correct instance.
    """
    await chancy.declare(Queue("low", executor=executor))

    ref = await chancy.push(Job.from_func(job_with_instance, queue="low"))
    job = await chancy.wait_for_job(ref)

    assert job.state == job.State.SUCCEEDED
    assert job.meta.get("received_instance") is True


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "executor",
    [
        "chancy.executors.asyncex.AsyncExecutor",
    ],
)
async def test_async_job_instance_kwarg(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task], executor: str
):
    """
    Test that async jobs requesting a QueuedJob kwarg receive the correct
    instance.
    """
    await chancy.declare(Queue("low", executor=executor))

    ref = await chancy.push(Job.from_func(async_job_with_instance, queue="low"))
    job = await chancy.wait_for_job(ref)

    assert job.state == job.State.SUCCEEDED
    assert job.meta.get("received_instance") is True



"""
Executors
=========

This package contains the built-in executors for Chancy. Executors are
responsible for managing the execution of jobs after they've been retrieved
from a queue.

Chancy comes with two built-in executors:

- :class:`chancy.executors.process.ProcessExecutor`: This executor will execute
    jobs in a separate process. This is the default executor.
- :class:`chancy.executors.asyncex.AsyncExecutor`: This executor will execute
    asyncio jobs in the same process as the worker.
- :class:`chancy.executors.thread.ThreadedExecutor`: This executor will execute
    jobs in a separate thread.

You can implement your own executor by subclassing the
:class:`chancy.executors.base.Executor` class and implementing the
:meth:`~chancy.executors.base.Executor.push` method.
"""



import abc
import typing


class RouteT(typing.TypedDict):
    """
    A type hint for a route.
    """

    path: str
    endpoint: typing.Callable
    methods: str | None
    name: str | None


class ApiPlugin(abc.ABC):
    """
    A plugin that provides additional API endpoints.
    """

    @abc.abstractmethod
    def name(self) -> str:
        """
        Get the name of the plugin.
        """

    def routes(self) -> list[RouteT]:
        """
        Get a list of routes to add to the API.
        """
        return []



from datetime import datetime, timezone, timedelta

from psycopg import sql
from psycopg import AsyncConnection

from chancy.plugin import Plugin, PluginScope
from chancy.app import Chancy
from chancy.worker import Worker


class Leadership(Plugin):
    """
    A plugin that manages cluster leadership.

    Leadership is used to ensure some plugins are run by only one worker at a
    time. For example, we wouldn't want every worker running database pruning
    every 60 seconds, as that would be immensely wasteful.

    This plugin is responsible for ensuring that only one worker is the leader
    at any given time. This plugin uses the {prefix}_leader table to store
    leadership information, which will only ever contain at most 1 row, the
    current leader.

    :param poll_interval: The number of seconds between leadership poll
                          intervals.
    :param timeout: The number of seconds before a worker is considered to have
                    lost leadership.
    """

    def __init__(self, poll_interval: int = 60, timeout: int = 60 * 2):
        super().__init__()
        self.poll_interval = poll_interval
        self.timeout = timeout

        if self.timeout < self.poll_interval:
            raise ValueError(
                "The timeout must be greater than the poll interval. Typically,"
                " the timeout should be at least twice the poll interval."
            )

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    async def run(self, worker: Worker, chancy: Chancy):
        prune_q = sql.SQL(
            """
            DELETE FROM {table} WHERE expires_at < %(now)s
            """
        ).format(table=sql.Identifier(f"{chancy.prefix}leader"))
        upsert_q = sql.SQL(
            """
            INSERT INTO {table} (id, worker_id, expires_at)
            VALUES (1, %(worker_id)s, %(expires_at)s)
            ON CONFLICT (id) DO UPDATE SET expires_at = %(expires_at)s
            """
        ).format(table=sql.Identifier(f"{chancy.prefix}leader"))
        insert_q = sql.SQL(
            """
            INSERT INTO {table} (id, worker_id, expires_at)
            VALUES (1, %(worker_id)s, %(expires_at)s) ON CONFLICT DO NOTHING
            """
        ).format(table=sql.Identifier(f"{chancy.prefix}leader"))

        # Trigger an immediate leadership check when a worker cleanly shuts
        # down.
        worker.hub.on(
            "worker.stopped",
            lambda worker_id: self.wakeup_signal.set(),
        )

        while await self.sleep(self.poll_interval):
            now = datetime.now(tz=timezone.utc)
            expires = now + timedelta(seconds=self.timeout)

            async with chancy.pool.connection() as conn:
                conn: AsyncConnection
                async with conn.cursor() as cur:
                    async with conn.transaction():
                        await cur.execute(prune_q, {"now": now})
                        if worker.is_leader.is_set():
                            await cur.execute(
                                upsert_q,
                                {
                                    "worker_id": worker.worker_id,
                                    "expires_at": expires,
                                },
                            )
                        else:
                            await cur.execute(
                                insert_q,
                                {
                                    "worker_id": worker.worker_id,
                                    "expires_at": expires,
                                },
                            )

                    is_leader = cur.rowcount == 1
                    was_leader = worker.is_leader.is_set()

                    if is_leader and was_leader:
                        chancy.log.debug(
                            f"Worker has renewed its leadership of the cluster"
                            f" until {expires}."
                        )
                        await chancy.notify(
                            cur,
                            "leadership.renewed",
                            {
                                "worker_id": worker.worker_id,
                            },
                        )
                    elif is_leader:
                        chancy.log.info(
                            f"Worker has become the leader of the cluster"
                            f" until {expires}."
                        )
                        worker.is_leader.set()
                        await chancy.notify(
                            cur,
                            "leadership.gained",
                            {
                                "worker_id": worker.worker_id,
                            },
                        )
                    elif was_leader:
                        chancy.log.info(
                            "Worker has lost leadership of the cluster."
                        )
                        worker.is_leader.clear()
                        await chancy.notify(
                            cur,
                            "leadership.lost",
                            {
                                "worker_id": worker.worker_id,
                            },
                        )



"""
Utilities for performing database migrations in a PostgreSQL database.

The functionality in this module is standalone and can be used with any
PostgreSQL database without any dependencies on the rest of the Chancy
project.
"""

import re
import abc
import importlib.resources

from psycopg import AsyncConnection, AsyncCursor
from psycopg import sql

VERSION_R = re.compile(r"v(\d+)\.py")


class MigrationError(Exception):
    """
    An error occurred while migrating the database schema.
    """


class Migration(abc.ABC):
    """
    A migration is a single unit of work that is applied to the database to
    bring it from one schema version to another.
    """

    @abc.abstractmethod
    async def up(self, migrator: "Migrator", cursor: AsyncCursor):
        pass

    @abc.abstractmethod
    async def down(self, migrator: "Migrator", cursor: AsyncCursor):
        pass


class Migrator:
    """
    A migrator is responsible for managing the database schema version and
    applying migrations to the database.

    :param key: A unique identifier for the application.
    :param migrations_package: The package where migrations are stored.
    :param prefix: A prefix to apply to all tables.
    """

    def __init__(self, key: str, migrations_package: str, *, prefix: str = ""):
        self.prefix = prefix
        self.key = key
        self.migrations_package = migrations_package

    def discover_all_migrations(self) -> list[Migration]:
        """
        Discovers all available migrations in the migrations package.

        Migrations are discovered by looking for classes that inherit from the
        `Migration` class.

        Migrations are sorted by their version number, which is the number at
        the beginning of the migration filename, ignoring the `v` prefix.

        For example, a migration file named `v1.py` would have a version number
        of 1.
        """
        migrations = []

        all_migrations = (
            resource
            for resource in importlib.resources.files(
                self.migrations_package
            ).iterdir()
            if resource.is_file()
        )

        for migration in all_migrations:
            if not VERSION_R.match(migration.name):
                continue

            module = importlib.import_module(
                f"{self.migrations_package}.{migration.name[:-3]}"
            )
            found_migration = False
            for name in dir(module):
                obj = getattr(module, name)
                if (
                    isinstance(obj, type)
                    and issubclass(obj, Migration)
                    and obj != Migration
                ):
                    if found_migration:
                        raise ValueError(
                            f"Multiple migrations found in {migration.name!r}"
                        )

                    migrations.append((int(migration.name[1:-3]), obj()))
                    found_migration = True

        return [migration for _, migration in sorted(migrations)]

    async def migrate(
        self, conn: AsyncConnection, to_version: int | None = None
    ) -> bool:
        """
        Migrate the database schema to the given version.

        This will migrate the database schema up or down as necessary to reach
        the given version. If `to_version` is less than the current schema
        version, the database will be migrated down. If `to_version` is greater
        than the current schema version, the database will be migrated up.

        If `to_version` is not provided, the database will be migrated to the
        highest available version.
        """
        migrations = self.discover_all_migrations()
        to_version = len(migrations) if to_version is None else to_version

        async with conn.cursor() as cursor:
            async with conn.transaction():
                current_version = await self.get_current_version(cursor)

                if current_version == to_version:
                    return False

                if to_version > len(migrations):
                    raise MigrationError(
                        f"Migration {to_version} does not exist for {self.key}"
                    )

                while current_version != to_version:
                    if current_version < to_version:
                        current_version += 1
                        await migrations[current_version - 1].up(self, cursor)
                        await self.set_current_version(cursor, current_version)
                    else:
                        await migrations[current_version - 1].down(self, cursor)
                        await self.set_current_version(
                            cursor, current_version - 1
                        )
                        current_version -= 1

        return True

    async def is_migration_required(self, cursor: AsyncCursor) -> bool:
        """
        Check if a newer schema version is available.
        """
        await self.upsert_version_table(cursor)
        current_version = await self.get_current_version(cursor)
        migrations = self.discover_all_migrations()
        return current_version < len(migrations)

    async def upsert_version_table(self, cursor: AsyncCursor):
        """
        Create the schema_version table if it doesn't exist.
        """
        await cursor.execute(
            sql.SQL(
                """
                CREATE TABLE IF NOT EXISTS {prefix} (
                    version_of VARCHAR(255) PRIMARY KEY,
                    version INT
                )
                """
            ).format(prefix=sql.Identifier(f"{self.prefix}schema_version"))
        )

    async def get_current_version(self, cursor: AsyncCursor) -> int:
        """
        Get the current schema version from the database.
        """
        await self.upsert_version_table(cursor)
        await cursor.execute(
            sql.SQL(
                "SELECT version FROM {prefix} WHERE version_of = %s"
            ).format(prefix=sql.Identifier(f"{self.prefix}schema_version")),
            [self.key],
        )
        result = await cursor.fetchone()
        return 0 if result is None else result[0]

    async def set_current_version(self, cursor: AsyncCursor, version: int):
        """
        Set the current schema version in the database.

        .. note::

            This does not perform any sanity checks nor does it run any
            migrations. It simply sets the version in the database.
        """
        await self.upsert_version_table(cursor)
        await cursor.execute(
            sql.SQL(
                """
                INSERT INTO {prefix} (version_of, version) VALUES
                (%s, %s) ON CONFLICT (version_of) DO UPDATE SET
                version = EXCLUDED.version
                """
            ).format(prefix=sql.Identifier(f"{self.prefix}schema_version")),
            [self.key, version],
        )



from psycopg import AsyncCursor
from psycopg import sql

from chancy.app import Chancy
from chancy.worker import Worker
from chancy.plugin import Plugin, PluginScope
from chancy.rule import SQLAble, JobRules
from chancy.utils import timed_block


class Pruner(Plugin):
    """
    A plugin that prunes stale data from the database.

    Jobs
    ----

    The pruner will never prune jobs that haven't been run yet ("pending"),
    or are currently being run ("running").

    You can use simple rules, or combine them using the `|` and `&` operators
    to create complex rules.

    For example, to prune jobs that are older than 60 seconds:

    .. code-block:: python

        Pruner(Pruner.Rules.Age() > 60)

    Or to prune jobs that are older than 60 seconds and are in the "default"
    queue:

    .. code-block:: python

        Pruner(Pruner.Rules.Queue() == "default" & (Pruner.Rules.Age() > 60))

    Or to prune jobs that are older than 60 seconds and are in the "default"
    queue, or instantly deleted if the job is `update_cache`:

    .. code-block:: python

        Pruner(
            (Pruner.Rules.Queue() == "default" & (Pruner.Rules.Age() > 60)) |
            Pruner.Rules.Job() == "update_cache"
        )

    By default, the pruner will run every 60 seconds and will remove up to
    10,000 jobs in a single run that have been completed for more than 60
    seconds.

    .. note::

        By default, only an Age rule will be covered by an index. If you use
        multiple rules, you may need to create additional indexes to improve
        performance on busy queues.

    :param rule: The rule that the pruner will use to match jobs.
    :param maximum_to_prune: The maximum number of jobs to prune in a single
                             run of the pruner.
    :param poll_interval: The interval in seconds between each run of the
                          pruner.
    """

    Rules = JobRules

    def __init__(
        self,
        rule: SQLAble = Rules.Age() > 60 * 60 * 24,
        *,
        maximum_to_prune: int = 10000,
        poll_interval: int = 60,
    ):
        super().__init__()
        self.rule = rule
        self.maximum_to_prune = maximum_to_prune
        self.poll_interval = poll_interval

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    async def run(self, worker: Worker, chancy: Chancy):
        while await self.sleep(self.poll_interval):
            await self.wait_for_leader(worker)
            async with chancy.pool.connection() as conn:
                async with conn.cursor() as cursor:
                    with timed_block() as chancy_time:
                        rows_removed = await self.prune(chancy, cursor)
                        chancy.log.info(
                            f"Pruner removed {rows_removed} row(s) from the"
                            f" database. Took {chancy_time.elapsed:.2f}"
                            f" seconds."
                        )

                        await chancy.notify(
                            cursor,
                            "pruner.removed",
                            {
                                "elapsed": chancy_time.elapsed,
                                "rows_removed": rows_removed,
                            },
                        )

            for plugin in chancy.plugins:
                rows = await plugin.cleanup(chancy)
                if rows is None:
                    continue

                chancy.log.info(
                    f"Plugin {plugin.__class__.__name__} removed {rows}"
                    f" row(s) from the database."
                )

    async def prune(self, chancy: Chancy, cursor: AsyncCursor) -> int:
        """
        Prune stale records from the database.

        :param chancy: The Chancy application.
        :param cursor: The database cursor to use for the operation.
        :return: The number of rows removed from the database
        """
        job_query = sql.SQL(
            """
            WITH jobs_to_prune AS (
                SELECT ctid
                FROM {table}
                WHERE state NOT IN ('pending', 'running')
                AND ({rule})
                LIMIT {maximum_to_prune}
            )
            DELETE FROM {table}
            WHERE ctid IN (SELECT ctid FROM jobs_to_prune)
            """
        ).format(
            table=sql.Identifier(f"{chancy.prefix}jobs"),
            rule=self.rule.to_sql(),
            maximum_to_prune=sql.Literal(self.maximum_to_prune),
        )

        await cursor.execute(job_query)
        return cursor.rowcount



import asyncio

import pytest

from chancy import Worker, Chancy, Job, Queue, QueuedJob


low = Queue("low", concurrency=1)
high = Queue("high", concurrency=1)


def job_to_run():
    return


def job_that_fails():
    raise ValueError("This job should fail.")


@pytest.mark.asyncio
async def test_multiple_queues(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task]
):
    """
    Ensure that jobs can be pushed & retrieved from multiple enabled queues.
    """
    await chancy.declare(low)
    await chancy.declare(high)

    ref_low = await chancy.push(Job.from_func(job_to_run, queue="low"))
    ref_high = await chancy.push(Job.from_func(job_to_run, queue="high"))

    job_low = await chancy.wait_for_job(ref_low)
    job_high = await chancy.wait_for_job(ref_high)

    assert job_low.state == QueuedJob.State.SUCCEEDED
    assert job_high.state == QueuedJob.State.SUCCEEDED


@pytest.mark.asyncio
async def test_queue_iteration(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task]
):
    """
    Ensure that we can retrieve and iterate over all the queues in the
    system using the Chancy object.
    """
    await chancy.declare(low)
    await chancy.declare(high)

    async for queue in chancy:
        assert queue in (low, high)


@pytest.mark.asyncio
async def test_queue_listing(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task]
):
    """
    Ensure that we can retrieve a list of all the queues in the system.
    """
    await chancy.declare(low)
    await chancy.declare(high)

    queues = await chancy.get_all_queues()
    assert len(queues) == 2
    assert low in queues
    assert high in queues


@pytest.mark.asyncio
async def test_get_single_queue(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task]
):
    """
    Ensure that we can retrieve a single queue from the system.
    """
    await chancy.declare(low)
    await chancy.declare(high)

    queue = await chancy.get_queue("low")
    assert queue == low

    queue = await chancy.get_queue("high")
    assert queue == high



from psycopg import AsyncCursor
from psycopg import sql

from chancy.migrate import Migration, Migrator


class V1Migration(Migration):
    async def up(self, migrator: Migrator, cursor: AsyncCursor):
        # Create workflows table
        await cursor.execute(
            sql.SQL(
                """
                CREATE TABLE {workflows} (
                    id UUID PRIMARY KEY,
                    name TEXT NOT NULL,
                    state TEXT NOT NULL,
                    created_at TIMESTAMPTZ DEFAULT NOW(),
                    updated_at TIMESTAMPTZ DEFAULT NOW()
                )
                """
            ).format(workflows=sql.Identifier(f"{migrator.prefix}workflows"))
        )

        # Create workflow_steps table
        await cursor.execute(
            sql.SQL(
                """
                CREATE TABLE {workflow_steps} (
                    id SERIAL PRIMARY KEY,
                    workflow_id UUID REFERENCES {workflows}(id)
                        ON DELETE CASCADE,
                    step_id TEXT NOT NULL,
                    job_data JSON NOT NULL,
                    dependencies JSON NOT NULL,
                    job_id UUID,
                    created_at TIMESTAMPTZ DEFAULT NOW(),
                    updated_at TIMESTAMPTZ DEFAULT NOW()
                )
                """
            ).format(
                workflow_steps=sql.Identifier(
                    f"{migrator.prefix}workflow_steps"
                ),
                workflows=sql.Identifier(f"{migrator.prefix}workflows"),
            )
        )

        # Create indexes
        await cursor.execute(
            sql.SQL(
                """
                CREATE INDEX {workflow_steps_workflow_id_idx} ON {workflow_steps} (workflow_id);
                CREATE INDEX {workflows_state_idx} ON {workflows} (state);
                CREATE UNIQUE INDEX {workflows_steps_unique_idx} ON {workflow_steps} (workflow_id, step_id);
                """
            ).format(
                workflow_steps=sql.Identifier(
                    f"{migrator.prefix}workflow_steps"
                ),
                workflows=sql.Identifier(f"{migrator.prefix}workflows"),
                workflow_steps_workflow_id_idx=sql.Identifier(
                    f"{migrator.prefix}workflow_steps_workflow_id_idx"
                ),
                workflow_steps_state_idx=sql.Identifier(
                    f"{migrator.prefix}workflow_steps_state_idx"
                ),
                workflows_state_idx=sql.Identifier(
                    f"{migrator.prefix}workflows_state_idx"
                ),
                workflows_steps_unique_idx=sql.Identifier(
                    f"{migrator.prefix}workflows_steps_unique_idx"
                ),
            )
        )

    async def down(self, migrator: Migrator, cursor: AsyncCursor):
        # Drop indexes
        await cursor.execute(
            sql.SQL(
                """
                DROP INDEX IF EXISTS {workflow_steps_workflow_id_idx};
                DROP INDEX IF EXISTS {workflow_steps_state_idx};
                DROP INDEX IF EXISTS {workflows_state_idx};
                """
            ).format(
                workflow_steps_workflow_id_idx=sql.Identifier(
                    f"{migrator.prefix}workflow_steps_workflow_id_idx"
                ),
                workflow_steps_state_idx=sql.Identifier(
                    f"{migrator.prefix}workflow_steps_state_idx"
                ),
                workflows_state_idx=sql.Identifier(
                    f"{migrator.prefix}workflows_state_idx"
                ),
            )
        )

        # Drop the workflow_steps table
        await cursor.execute(
            sql.SQL("DROP TABLE IF EXISTS {table}").format(
                table=sql.Identifier(f"{migrator.prefix}workflow_steps")
            )
        )

        # Drop the workflows table
        await cursor.execute(
            sql.SQL("DROP TABLE IF EXISTS {table}").format(
                table=sql.Identifier(f"{migrator.prefix}workflows")
            )
        )



from psycopg import AsyncCursor
from psycopg import sql

from chancy.migrate import Migration, Migrator


class V2Migration(Migration):
    async def up(self, migrator: Migrator, cursor: AsyncCursor):
        """
        Adds common indexes.
        """
        await cursor.execute(
            sql.SQL(
                """
                CREATE INDEX IF NOT EXISTS {idx_queue_state}
                ON {jobs} (queue, state);
                """
            ).format(
                jobs=sql.Identifier(f"{migrator.prefix}jobs"),
                idx_queue_state=sql.Identifier(
                    f"{migrator.prefix}idx_jobs_queue_state"
                ),
            )
        )

        await cursor.execute(
            sql.SQL(
                """
                CREATE INDEX IF NOT EXISTS {idx_jobs_fetch_prune} ON {jobs}
                (queue, state, attempts, scheduled_at, priority DESC, id DESC)
                WHERE state IN ('pending', 'retrying');
                """
            ).format(
                jobs=sql.Identifier(f"{migrator.prefix}jobs"),
                idx_jobs_fetch_prune=sql.Identifier(
                    f"{migrator.prefix}idx_jobs_fetch_prune"
                ),
            )
        )

    async def down(self, migrator: Migrator, cursor: AsyncCursor):
        await cursor.execute(
            sql.SQL(
                """
                DROP INDEX IF EXISTS {idx_queue_state};
                """
            ).format(
                idx_queue_state=sql.Identifier(
                    f"{migrator.prefix}idx_jobs_queue_state"
                )
            )
        )

        await cursor.execute(
            sql.SQL(
                """
                DROP INDEX IF EXISTS {idx_jobs_fetch_prune};
                """
            ).format(
                idx_jobs_fetch_prune=sql.Identifier(
                    f"{migrator.prefix}idx_jobs_fetch_prune"
                )
            )
        )



import json

import click
from click import Context

from chancy import Chancy, Job, Queue
from chancy.cli import run_async_command


@click.group(name="queue")
def queue_group():
    """
    Queue management commands.
    """
    pass


@queue_group.command()
@click.argument("job")
@click.option(
    "--queue",
    "-q",
    help="The queue to push the job to.",
    default="default",
)
@click.option(
    "--priority",
    "-p",
    help="The job's priority.",
    default=0,
    type=int,
)
@click.option(
    "--unique-key",
    "-u",
    help="The job's unique key.",
)
@click.option(
    "--max-attempts",
    "-m",
    help="The maximum number of attempts to make.",
    type=int,
)
@click.option(
    "--kwargs",
    "-k",
    help="A JSON blob containing the keyword arguments to pass to the job.",
)
@click.pass_context
@run_async_command
async def push(
    ctx: Context,
    job: str,
    queue: str,
    priority: int,
    unique_key: str,
    max_attempts: int,
    kwargs: str | None = None,
):
    """
    Push a job to the default queue.
    """
    chancy: Chancy = ctx.obj["app"]

    if kwargs is not None:
        kwargs = json.loads(kwargs)

    async with chancy:
        await chancy.push(
            Job(
                func=job,
                queue=queue,
                priority=priority,
                unique_key=unique_key,
                max_attempts=max_attempts,
                kwargs=kwargs,
            )
        )


@queue_group.command("list")
@click.pass_context
@run_async_command
async def list_queues(ctx: click.Context):
    """
    List all the queues.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        queues = await chancy.get_all_queues()

        for queue in queues:
            click.echo(f"{queue!r}")


@queue_group.command("declare")
@click.argument("name")
@click.option(
    "--concurrency",
    "-c",
    default=1,
    type=int,
    help="The number of simultaneous jobs to process.",
)
@click.option(
    "--executor",
    "-e",
    default="chancy.executors.process.ProcessExecutor",
    help="The executor to use.",
)
@click.option(
    "--rate-limit", "-r", type=int, help="The global queue rate limit."
)
@click.option(
    "--rate-limit-window",
    "-w",
    type=int,
    help="The global queue rate limit window.",
)
@click.option(
    "--polling-interval",
    "-p",
    help="The interval to poll the queue for new jobs.",
    type=int,
    default=5,
)
@click.option(
    "--tags",
    "-t",
    help="Extra tags to apply to the queue.",
    multiple=True,
)
@click.pass_context
@run_async_command
async def declare_queue(
    ctx: click.Context,
    name: str,
    concurrency: int | None,
    executor: str | None,
    rate_limit: int | None,
    rate_limit_window: int | None,
    polling_interval: int | None,
    tags: list[str] | None,
):
    """
    Declare a new queue.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        await chancy.declare(
            Queue(
                name,
                concurrency=concurrency,
                executor=executor,
                rate_limit=rate_limit,
                rate_limit_window=rate_limit_window,
                polling_interval=polling_interval,
                tags=set(tags) if tags else {r".*"},
            )
        )


@queue_group.command("delete")
@click.argument("name")
@click.pass_context
@run_async_command
async def delete_queue(ctx: click.Context, name: str):
    """
    Delete a queue.
    """
    chancy: Chancy = ctx.obj["app"]

    if click.confirm(f"Are you sure you want to delete the queue {name}?"):
        async with chancy:
            await chancy.delete_queue(name)



import asyncio
import inspect
from typing import Callable, Awaitable


class Hub:
    """
    An event hub for registering and emitting events.
    """

    def __init__(self):
        self._handlers: dict[str, list[Callable[..., Awaitable[None]]]] = {}
        self._waiters: dict[str, list[asyncio.Future]] = {}

    def on(self, event: str, f: Callable[..., Awaitable[None]]):
        """
        Register a handler for an event.
        """
        self._handlers.setdefault(event, []).append(f)

    def remove(self, event: str, handler: Callable[..., Awaitable[None]]):
        """
        Remove a handler from an event.
        """
        try:
            self._handlers.get(event, []).remove(handler)
        except ValueError:
            pass

    async def emit(self, event: str, body):
        """
        Emit an event with the given body.
        """
        for handler in self._handlers.get(event, []):
            # Check if the handler is async
            if inspect.iscoroutinefunction(handler):
                await handler(body)
            else:
                handler(body)

        # Emit wildcard handlers.
        if "*" in self._handlers:
            for handler in self._handlers["*"]:
                if inspect.iscoroutinefunction(handler):
                    await handler(event, body)
                else:
                    handler(event, body)

        # Resolve any waiting futures
        waiters = self._waiters.get(event, [])
        for waiter in waiters:
            if not waiter.done():
                waiter.set_result(body)
        self._waiters[event] = [w for w in waiters if not w.done()]

    def off(self, event: str, callback: Callable[..., Awaitable[None]]):
        """
        Remove a callback from an event.

        :param event: The event to remove the callback from.
        :param callback: The callback to remove.
        """
        self._handlers[event] = [
            h for h in self._handlers[event] if h != callback
        ]

    async def wait_for(self, event: str, timeout: float | None = None):
        """
        Wait for a specific event to occur.

        :param event: The event to wait for.
        :param timeout: The maximum time to wait for the event (in seconds).
        :return: The data associated with the event.
        :raises TimeoutError: If the timeout is reached before the event occurs.
        """
        future = asyncio.get_running_loop().create_future()
        self._waiters.setdefault(event, []).append(future)

        try:
            return await asyncio.wait_for(future, timeout)
        except TimeoutError:
            self._waiters[event].remove(future)
            raise
        finally:
            if not future.done():
                future.cancel()



import pytest

from chancy import Chancy, QueuedJob, Job, Queue


def job_to_run():
    return


@pytest.mark.asyncio
async def test_basic_job(chancy_just_app: Chancy):
    """
    Simply test that we can push a job, and it runs successfully.
    """
    await chancy_just_app.migrate()
    await chancy_just_app.declare(Queue("default"))
    ref = await chancy_just_app.push(Job.from_func(job_to_run))
    job = await chancy_just_app.get_job(ref)
    assert job.state == QueuedJob.State.PENDING
    await chancy_just_app.migrate(to_version=0)
    await chancy_just_app.pool.close()



import asyncio

import pytest

from chancy import Worker, Chancy, Job, Queue, QueuedJob


def job_to_run():
    return


def job_that_fails():
    raise ValueError("This job should fail.")


@pytest.mark.asyncio
async def test_basic_job(chancy: Chancy, worker: tuple[Worker, asyncio.Task]):
    """
    Simply test that we can push a job, and it runs successfully.
    """
    await chancy.declare(
        Queue(
            "default",
            concurrency=1,
        ),
        upsert=True,
    )
    ref = await chancy.push(Job.from_func(job_to_run))
    job = await chancy.wait_for_job(ref)
    assert job.state == QueuedJob.State.SUCCEEDED


@pytest.mark.asyncio
async def test_failing_job(chancy: Chancy, worker: tuple[Worker, asyncio.Task]):
    """
    Test that a job that fails will be marked as failed.
    """
    await chancy.declare(
        Queue(
            "default",
            concurrency=1,
        ),
        upsert=True,
    )
    ref = await chancy.push(Job.from_func(job_that_fails))
    job = await chancy.wait_for_job(ref)
    assert job.state == QueuedJob.State.FAILED



import abc
import enum
import asyncio
import typing

from chancy.job import QueuedJob
from chancy.migrate import Migrator

if typing.TYPE_CHECKING:
    from chancy.app import Chancy
    from chancy.worker import Worker


class PluginScope(enum.Enum):
    #: The plugin is a general worker plugin.
    WORKER = "worker"


class Plugin(abc.ABC):
    """
    Base class for all plugins.

    Plugins are used to extend the functionality of the worker. When a worker
    starts, it will call :meth:`run` on all plugins that have a scope that
    matches the worker's scope.
    """

    def __init__(self):
        #: An asyncio.Event that can be used to wake up the plugin if it's
        #: sleeping.
        self.wakeup_signal = asyncio.Event()

    @classmethod
    @abc.abstractmethod
    def get_scope(cls) -> PluginScope:
        """
        Get the scope of this plugin.
        """

    async def run(self, worker: "Worker", chancy: "Chancy"):
        """
        Runs the plugin.

        This function can and should run indefinitely, as it will be cancelled
        when the worker is stopped.
        """

    def migrate_package(self) -> str | None:
        """
        Get the package name that contains the migration scripts for this
        plugin, if it has any.
        """

    def migrate_key(self) -> str | None:
        """
        Get the migration key for this plugin, if it has any.
        """

    async def migrate(self, chancy: "Chancy", *, to_version: int | None = None):
        """
        Migrate the database to the latest schema version.

        If `to_version` is provided, the database will be migrated to that
        specific version, up or down as necessary.
        """
        migrator = self.migrator(chancy)
        if migrator is None:
            return

        async with chancy.pool.connection() as conn:
            await migrator.migrate(conn, to_version=to_version)

    def migrator(self, chancy: "Chancy") -> Migrator | None:
        """
        Get a migrator for this plugin, if it has any migration scripts.
        """
        key = self.migrate_key()
        if key is None:
            return

        package = self.migrate_package()
        if package is None:
            return

        return Migrator(key, package, prefix=chancy.prefix)

    async def sleep(self, seconds: int) -> bool:
        """
        Sleep for a specified number of seconds, but allow the plugin to be
        woken up early.
        """
        wakeup = asyncio.create_task(self.wakeup_signal.wait())

        done, pending = await asyncio.wait(
            [wakeup],
            timeout=seconds,
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in pending:
            task.cancel()

        self.wakeup_signal.clear()
        return True

    @staticmethod
    async def wait_for_leader(worker: "Worker") -> None:
        """
        Wait until the worker running this plugin is the leader.
        """
        return await worker.is_leader.wait()

    def wake_up(self):
        """
        Wake up the plugin if it's sleeping.
        """
        self.wakeup_signal.set()

    async def cleanup(self, chancy: "Chancy") -> int | None:
        """
        Clean up any resources used by the plugin.

        Should return either None, if no work was done, or the number of
        rows cleaned up.

        .. note::

            Normally, you don't need to call this yourself. The Pruner plugin
            will call the cleanup method of all other registered plugins.
        """

    def api_plugin(self) -> str | None:
        """
        If this plugin has an associated API component, returns the import
        string for the plugin.
        """

    async def on_job_completed(
        self,
        job: QueuedJob,
        worker: "Worker",
        *,
        exc: Exception | None = None,
    ) -> QueuedJob:
        """
        Called after a job is completed (successfully or otherwise) and before
        the QueuedJob is updated in the database.

        If an exception occurred during the job, `exc` will be the exception
        instance instead of ``None``.

        The passed job is immutable - to modify it, return a new QueuedJob
        object with the desired changes.
        """
        raise NotImplementedError()

    def __repr__(self):
        return f"<{self.__class__.__name__}()>"
